"""
Custom lazy Greedy Best-First Search (GBFS) with customization points explained.

Tyr also provides an off-the-shelf implementation, which is almost identical
to the astar_eager.py example. Please familiarize yourself with astar_eager.py
before exploring this example.

Example usage (run from the repository root):

    python3 python/examples/planning/gbfs_lazy.py \
        -d data/gripper/domain.pddl \
        -p data/gripper/test_problem.pddl

Author: Dominik Drexler (dominik.drexler@liu.se)
"""

import argparse

from pathlib import Path
from queue import PriorityQueue
from dataclasses import dataclass
from enum import Enum

from pytyr.formalism.planning import ParserOptions, Parser
from pytyr.planning import SearchStatus
from pytyr.planning.lifted import StateIndex, SearchResult, SuccessorGenerator, Heuristic, FFHeuristic, PruningStrategy, GoalStrategy, TaskGoalStrategy, State, Node, LabeledNode, Plan, Task


class SearchNodeStatus(Enum):
        NEW = 0
        OPEN = 1
        CLOSED = 2
        DEADEND = 3
        GOAL = 4


@dataclass
class SearchNode:
    g_value : float
    parent : StateIndex | None
    status : SearchNodeStatus


def backtrack_plan(goal_node : Node, goal_search_node : SearchNode, search_nodes : dict[StateIndex, SearchNode], successor_generator : SuccessorGenerator):
    """
    Backtracks from the goal search node to compute the plan inducing state trajectory,
    followed by computing the ground actions connecting subsequent state in the trajectory.
    """
    cur_state_index = goal_node.get_state().get_index()
    backward_state_trajectory = []

    while cur_state_index is not None:
        backward_state_trajectory.append(cur_state_index)
        search_node = search_nodes.get(cur_state_index)
        cur_state_index = search_node.parent

    forward_state_trajectory = list(reversed(backward_state_trajectory))

    assert(forward_state_trajectory)

    start_node = successor_generator.get_node(forward_state_trajectory[0])
    labeled_succ_nodes : list[LabeledNode] = []
    
    for i in range(len(forward_state_trajectory) - 1):
        state_index = forward_state_trajectory[i]
        node = successor_generator.get_node(state_index)

        for labeled_succ_node in successor_generator.get_labeled_successor_nodes(node):
            succ_node = labeled_succ_node.node
            succ_state = succ_node.get_state()
            succ_state_index = succ_state.get_index()

            if succ_state_index == forward_state_trajectory[i + 1]:
                labeled_succ_nodes.append(labeled_succ_node)

    return Plan(start_node, labeled_succ_nodes)


def find_solution(task : Task, heuristic : Heuristic) -> SearchResult: 
    successor_generator = SuccessorGenerator(task)
    state_repository = successor_generator.get_state_repository()

    goal_strategy = TaskGoalStrategy(task)

    search_result = SearchResult()

    if not goal_strategy.is_static_goal_satisfied():
        search_result.goal_node = None
        search_result.plan = None
        search_result.status = SearchStatus.UNSOLVABLE
        return search_result

    initial_node = successor_generator.get_initial_node()
    initial_state = initial_node.get_state()

    if goal_strategy.is_dynamic_goal_satisfied(initial_state):
        search_result.goal_node = initial_node
        search_result.plan = Plan(initial_node)
        search_result.status = SearchStatus.SOLVED
        return search_result

    initial_h_value = heuristic.evaluate(initial_state)
    initial_g_value = initial_node.get_metric()
    initial_state_index = initial_state.get_index()

    search_nodes : dict[StateIndex, SearchNode] = dict()
    search_nodes[initial_state_index] = SearchNode(initial_g_value, None, SearchNodeStatus.NEW)

    queue = PriorityQueue()
    # Note: we insert cheap state indices to allow memory to be returned to the pool.
    # We can always retrieve a state by calling state_repository.get_registered_state(i) for a StateIndex i.
    queue.put((initial_h_value, initial_g_value, initial_state_index))

    while not queue.empty():
        h_value, g_value, state_index = queue.get()
        state = state_repository.get_registered_state(state_index)
        node = Node(state, g_value)
        search_node = search_nodes.get(state_index)

        state_h_value = heuristic.evaluate(state)
        preferred_actions = heuristic.get_preferred_actions()

        search_node.status = SearchNodeStatus.CLOSED

        if state_h_value == float("inf"):
            search_node.status = SearchNodeStatus.DEADEND
            continue

        for labeled_succ_node in successor_generator.get_labeled_successor_nodes(node):
            succ_node = labeled_succ_node.node
            succ_state = succ_node.get_state()
            succ_g_value = succ_node.get_metric()
            succ_state_index = succ_state.get_index()

            if succ_state_index not in search_nodes:
                search_nodes[succ_state_index] = SearchNode(succ_g_value, state_index, SearchNodeStatus.NEW)

            succ_search_node : SearchNode = search_nodes.get(succ_state_index)
        
            if not succ_search_node.status == SearchNodeStatus.NEW:
                continue

            succ_search_node.parent = state_index

            action = labeled_succ_node.label
            if action in preferred_actions:
                pass  # the heuristic marked this labeled successor node as preferred

            if goal_strategy.is_dynamic_goal_satisfied(succ_state):
                search_result.goal_node = succ_node
                search_result.plan = backtrack_plan(succ_node, succ_search_node, search_nodes, successor_generator)
                search_result.status = SearchStatus.SOLVED
                return search_result
            
            succ_search_node.status = SearchNodeStatus.OPEN
            queue.put((state_h_value, succ_g_value, succ_state_index))

            
    search_result.goal_node = None
    search_result.plan = None
    search_result.status = SearchStatus.EXHAUSTED
    return search_result


def main():
    arg_parser = argparse.ArgumentParser(description="GBFS Lazy Search.")
    arg_parser.add_argument("-d", "--domain-filepath", type=Path, required=True, help="Path to a PDDL domain file.")
    arg_parser.add_argument("-p", "--task-filepath", type=Path, required=True, help="Path to PDDL task file.")
    args = arg_parser.parse_args()

    domain_filepath : Path = args.domain_filepath
    task_filepath : Path = args.task_filepath

    parser_options = ParserOptions()
    parser = Parser(domain_filepath, parser_options)
    lifted_task = Task(parser.parse_task(task_filepath, parser_options))
    heuristic = FFHeuristic(lifted_task)

    search_result = find_solution(lifted_task, heuristic)

    print("Search status:", search_result.status)

    plan = search_result.plan

    if plan is not None:
        print(f"Found plan with length {plan.get_length()} and cost {plan.get_cost()}")
        print(plan)
    else:
        print("No solution was found.")



if __name__ == "__main__":
    main()



from pytyr.formalism.planning import ParserOptions, Parser

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent.parent.parent).absolute()


def test_pddl_parser():
    """ Test parsing and translation of a PDDL domain and problem file.
    """
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "test_problem.pddl")
    parser_options = ParserOptions()
    parser = Parser(domain_filepath, parser_options)

    domain = parser.get_domain()
    task = parser.parse_task(problem_filepath, parser_options)

    assert(domain.get_domain() == task.get_domain().get_domain())


# Import all classes for better IDE support

from pytyr.pytyr.planning.ground import (
    Task,
    State,
    StateIndex,
    Node,
    LabeledNode,
    Plan,
    AxiomEvaluator,
    StateRepository,
    SuccessorGenerator,
    SearchResult,
    GoalStrategy,
    TaskGoalStrategy,
    PruningStrategy,
    Heuristic,
    BlindHeuristic,
)

from . import (
    astar_eager as astar_eager,
)

from . import (
    gbfs_lazy as gbfs_lazy,
)



# Import all classes for better IDE support

from pytyr.pytyr.planning.lifted import (
    Task,
    State,    
    StateIndex,
    Node,    
    LabeledNode,
    Plan,
    AxiomEvaluator,
    StateRepository,
    SuccessorGenerator,
    SearchResult,
    GoalStrategy,
    TaskGoalStrategy,
    PruningStrategy,
    Heuristic,
    BlindHeuristic,
    MaxHeuristic,
    AddHeuristic,
    FFHeuristic,
)

from . import (
    astar_eager as astar_eager,
)

from . import (
    gbfs_lazy as gbfs_lazy,
)




# Import all classes for better IDE support

from pytyr.pytyr.formalism.planning import (
    # Core symbols
    Object,
    ObjectIndex,
    ObjectBuilder,
    Binding,
    BindingIndex,
    BindingBuilder,
    Variable,
    VariableIndex,
    VariableBuilder,
    ParameterIndex,
    Term,

    StaticPredicateBinding,
    StaticPredicateBindingIndex,
    FluentPredicateBinding,
    FluentPredicateBindingIndex,
    DerivedPredicateBinding,
    DerivedPredicateBindingIndex,
    StaticFunctionBinding,
    StaticFunctionBindingIndex,
    FluentFunctionBinding,
    FluentFunctionBindingIndex,
    AuxiliaryFunctionBinding,
    AuxiliaryFunctionBindingIndex,
    ActionBinding,
    ActionBindingIndex,
    AxiomBinding,
    AxiomBindingIndex,

    # Predicates
    StaticPredicate,
    StaticPredicateIndex,
    StaticPredicateBuilder,
    FluentPredicate,
    FluentPredicateIndex,
    FluentPredicateBuilder,
    DerivedPredicate,
    DerivedPredicateIndex,
    DerivedPredicateBuilder,

    # Atoms
    StaticAtom,
    StaticAtomIndex,
    FluentAtom,
    FluentAtomIndex,
    DerivedAtom,
    DerivedAtomIndex,
    StaticGroundAtom,
    StaticGroundAtomIndex,
    StaticGroundAtomBuilder,
    FluentGroundAtom,
    FluentGroundAtomIndex,
    FluentGroundAtomBuilder,
    DerivedGroundAtom,
    DerivedGroundAtomIndex,
    DerivedGroundAtomBuilder,

    # Literals
    StaticLiteral,
    StaticLiteralIndex,
    FluentLiteral,
    FluentLiteralIndex,
    DerivedLiteral,
    DerivedLiteralIndex,
    StaticGroundLiteral,
    StaticGroundLiteralIndex,
    FluentGroundLiteral,
    FluentGroundLiteralIndex,
    DerivedGroundLiteral,
    DerivedGroundLiteralIndex,
    FluentFDRVariable,
    FluentFDRVariableIndex,
    FDRValue,
    FluentFDRFact,

    # Functions
    StaticFunction,
    StaticFunctionIndex,
    FluentFunction,
    FluentFunctionIndex,
    AuxiliaryFunction,
    AuxiliaryFunctionIndex,

    # FunctionTerms
    StaticFunctionTerm,
    StaticFunctionTermIndex,
    FluentFunctionTerm,
    FluentFunctionTermIndex,
    AuxiliaryFunctionTerm,
    AuxiliaryFunctionTermIndex,
    StaticGroundFunctionTerm,
    StaticGroundFunctionTermIndex,
    FluentGroundFunctionTerm,
    FluentGroundFunctionTermIndex,
    AuxiliaryGroundFunctionTerm,
    AuxiliaryGroundFunctionTermIndex,

    # FunctionTermValues
    StaticGroundFunctionTermValue,
    StaticGroundFunctionTermValueIndex,
    FluentGroundFunctionTermValue,
    FluentGroundFunctionTermValueIndex,
    AuxiliaryGroundFunctionTermValue,
    AuxiliaryGroundFunctionTermValueIndex,

    # Operators
    UnaryOperatorSub,
    UnaryOperatorSubIndex,
    BinaryOperatorAdd,
    BinaryOperatorAddIndex,
    BinaryOperatorSub,
    BinaryOperatorSubIndex,
    BinaryOperatorMul,
    BinaryOperatorMulIndex,
    BinaryOperatorDiv,
    BinaryOperatorDivIndex,
    BinaryOperatorEq,
    BinaryOperatorEqIndex,
    BinaryOperatorNe,
    BinaryOperatorNeIndex,
    BinaryOperatorLe,
    BinaryOperatorLeIndex,
    BinaryOperatorLt,
    BinaryOperatorLtIndex,
    BinaryOperatorGe,
    BinaryOperatorGeIndex,
    BinaryOperatorGt,
    BinaryOperatorGtIndex,
    MultiOperatorAdd,
    MultiOperatorAddIndex,
    MultiOperatorMul,
    MultiOperatorMulIndex,
    ArithmeticOperator,
    BooleanOperator,
    FunctionExpression,

    # Condition
    ConjunctiveCondition,
    ConjunctiveConditionIndex,

    # Effects
    FluentNumericEffectAssign,
    FluentNumericEffectAssignIndex,
    FluentNumericEffectIncrease,
    FluentNumericEffectIncreaseIndex,
    FluentNumericEffectDecrease,
    FluentNumericEffectDecreaseIndex,
    FluentNumericEffectScaleUp,
    FluentNumericEffectScaleUpIndex,
    FluentNumericEffectScaleDown,
    FluentNumericEffectScaleDownIndex,
    AuxiliaryNumericEffectIncrease,
    AuxiliaryNumericEffectIncreaseIndex,
    FluentNumericEffectOperator,
    AuxiliaryNumericEffectOperator,
    ConjunctiveEffect,
    ConjunctiveEffectIndex,
    ConditionalEffect,
    ConditionalEffectIndex,

    # Action
    Action,
    ActionIndex,

    # Action
    Axiom,
    AxiomIndex,

    # Ground Operators
    GroundBinaryOperatorAdd,
    GroundBinaryOperatorAddIndex,
    GroundBinaryOperatorSub,
    GroundBinaryOperatorSubIndex,
    GroundBinaryOperatorMul,
    GroundBinaryOperatorMulIndex,
    GroundBinaryOperatorDiv,
    GroundBinaryOperatorDivIndex,
    GroundBinaryOperatorEq,
    GroundBinaryOperatorEqIndex,
    GroundBinaryOperatorNe,
    GroundBinaryOperatorNeIndex,
    GroundBinaryOperatorLe,
    GroundBinaryOperatorLeIndex,
    GroundBinaryOperatorLt,
    GroundBinaryOperatorLtIndex,
    GroundBinaryOperatorGe,
    GroundBinaryOperatorGeIndex,
    GroundBinaryOperatorGt,
    GroundBinaryOperatorGtIndex,
    GroundMultiOperatorAdd,
    GroundMultiOperatorAddIndex,
    GroundMultiOperatorMul,
    GroundMultiOperatorMulIndex,
    GroundArithmeticOperator,
    GroundBooleanOperator,
    GroundFunctionExpression,

    # GroundCondition
    GroundConjunctiveCondition,
    GroundConjunctiveConditionIndex,

    # Ground NumericEffects
    FluentGroundNumericEffectAssign,
    FluentGroundNumericEffectAssignIndex,
    FluentGroundNumericEffectIncrease,
    FluentGroundNumericEffectIncreaseIndex,
    FluentGroundNumericEffectDecrease,
    FluentGroundNumericEffectDecreaseIndex,
    FluentGroundNumericEffectScaleUp,
    FluentGroundNumericEffectScaleUpIndex,
    FluentGroundNumericEffectScaleDown,
    FluentGroundNumericEffectScaleDownIndex,
    AuxiliaryGroundNumericEffectIncrease,
    AuxiliaryGroundNumericEffectIncreaseIndex,
    FluentGroundNumericEffectOperator,
    AuxiliaryGroundNumericEffectOperator,
    GroundConjunctiveEffect,
    GroundConjunctiveEffectIndex,
    GroundConditionalEffect,
    GroundConditionalEffectIndex,

    # GroundAction
    GroundAction,
    GroundActionIndex,

    # GroundAction
    GroundAxiom,
    GroundAxiomIndex,

    # Tasks
    Minimize,
    Maximize,
    Metric,
    MetricIndex,
    Domain,
    DomainIndex,
    LiftedTask,
    LiftedTaskIndex,
    GroundTask,
    GroundTaskIndex,
    PlanningDomain,
    PlanningTask,
    PlanningFDRTask,

    # Parser
    ParserOptions,
    Parser,

    # Repository,
    Repository,

    # FDRContext
    BinaryFDRContext,
    GeneralFDRContext,
)


"""
Off-the-shelf A* eager with customization points explained.

Example usage (run from the repository root):

    python3 python/examples/planning/astar_eager.py \
        -d data/gripper/domain.pddl \
        -p data/gripper/test_problem.pddl

Author: Dominik Drexler (dominik.drexler@liu.se)
"""

import argparse

from pathlib import Path

from pytyr.formalism.planning import ParserOptions, Parser, GroundConjunctiveCondition, GroundAction
# Note: we can easily switch between lifted and ground planning by swapping the submodule, 
# e.g., from pytyr.planning.ground import ..., and from pytyr.planning.ground.astar_eager import ...
# However, some heuristics, like MaxHeuristic might not be available yet.
from pytyr.planning.lifted import Task, SuccessorGenerator, Heuristic, MaxHeuristic, PruningStrategy, GoalStrategy, TaskGoalStrategy, State, Node, LabeledNode, Plan, Task
from pytyr.planning.lifted.astar_eager import Options, EventHandler, DefaultEventHandler, find_solution

# Lazy Greedy Best-First Search (GBFS) exposes an almost identical interface,
# although some options and EventHandler events differ slightly.
# To switch the search algorithm, typically only the following import needs to be changed:
#
# from pytyr.planning.lifted.gbfs_lazy import Options, EventHandler, DefaultEventHandler, find_solution


class CustomHeuristic(Heuristic):
    """
    CustomHeuristic derives from Heuristic to implement custom heuristics.
    """
    def set_goal(self, goal : GroundConjunctiveCondition) -> None:
        pass

    def evaluate(self, state : State) -> float:
        pass

    def get_preferred_actions(self) -> set[GroundAction]:
        pass


class CustomPruningStrategy(PruningStrategy):
    """
    CustomPruningStrategy derives from PruningStrategy to implement custom strategies to prune states during search.

    A prominent example is Iterative Width (IW) that prunes state in a breadth-first (BrFS) search based on a state novelty criteria.
    """
    def should_prune_state(self, state: State) -> bool: 
        """ Is checked for the initial state. """
        pass

    def should_prune_successor_state(self, state: State, succ_state: State, is_new_succ_state: bool) -> bool:
        """ Is checked for every generated state. """
        pass


class CustomGoalStrategy(GoalStrategy):
    """
    CustomGoalStrategy derives from GoalStrategy to implement custom strategies to terminate search upon reaching a goal state.

    A prominent example is Serialized Iterative Width (SIW) that repeateadly calls IW to greedily achieve one goal atom at a time.
    """
    def is_static_goal_satisfied(self) -> bool: 
        """ Is checked before running a search. """
        pass

    def is_dynamic_goal_satisfied(self, state: State) -> bool: 
        """ Is checked for every generated state. """
        pass


class CustomEventHandler(EventHandler):
    """
    CustomEventHandler derives from EventHandler to implement custom ways to react on events triggered during an A* eager search.

    We can use it for instance to create a search tree, or even a complete state space 
    when running A* with a blind heuristic and a goal strategy that always returns False.
    We can also use it to collect custom search statistics.
    """
    def on_expand_node(self, node: Node) -> None: 
        pass

    def on_expand_goal_node(self, node: Node) -> None:
        pass

    def on_generate_node(self, labeled_succ_node: LabeledNode) -> None:
        pass

    def on_generate_node_relaxed(self, labeled_succ_node: LabeledNode) -> None:
        pass

    def on_generate_node_not_relaxed(self, labeled_succ_node: LabeledNode) -> None:
        pass

    def on_close_node(self, node: Node) -> None:
        pass

    def on_prune_node(self, node: Node) -> None:
        pass

    def on_start_search(self, node: Node, f_value: float) -> None:
        pass

    def on_finish_f_layer(self, f_value: float) -> None:
        pass

    def on_end_search(self) -> None:
        pass

    def on_solved(self, plan: Plan) -> None:
        pass

    def on_unsolvable(self) -> None:
        pass

    def on_exhausted(self) -> None:
        pass

    
def main():
    arg_parser = argparse.ArgumentParser(description="A* Eager Search.")
    arg_parser.add_argument("-d", "--domain-filepath", type=Path, required=True, help="Path to a PDDL domain file.")
    arg_parser.add_argument("-p", "--task-filepath", type=Path, required=True, help="Path to PDDL task file.")
    args = arg_parser.parse_args()

    domain_filepath : Path = args.domain_filepath
    task_filepath : Path = args.task_filepath

    parser_options = ParserOptions()
    parser = Parser(domain_filepath, parser_options)
    lifted_task = Task(parser.parse_task(task_filepath, parser_options))
    heuristic = MaxHeuristic(lifted_task)
    successor_generator = SuccessorGenerator(lifted_task)

    options = Options()
    options.num_threads = 2                                # Lifted search is parallelized but only useful on large tasks.
    options.event_handler = DefaultEventHandler(0)         # Collects and prints statistics. If verbosity >= 2, then also prints labeled nodes.
    options.goal_strategy = TaskGoalStrategy(lifted_task)  # Terminates the search when reaching a state that satisfies the task's goal.
    options.pruning_strategy = PruningStrategy()           # Never prunes

    search_result = find_solution(lifted_task, successor_generator, heuristic, options)
 
    print("Search status:", search_result.status)

    plan = search_result.plan

    if plan is not None:
        print(f"Found plan with length {plan.get_length()} and cost {plan.get_cost()}")
        print(plan)
    else:
        print("No solution was found.")


if __name__ == "__main__":
    main()


import os
import sys
import subprocess
import multiprocessing
import shutil

from pathlib import Path

from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext


__version__ = "0.0.4"
HERE = Path(__file__).resolve().parent


# A CMakeExtension needs a sourcedir instead of a file list.
# The name must be the _single_ output extension from the CMake build.
# If you need multiple extensions, see scikit-build.
class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = Path(os.path.abspath(sourcedir))


def get_num_jobs():
    return int(os.environ.get("TYR_JOBS", multiprocessing.cpu_count()))

def get_header_instantiation():
    return os.environ.get("TYR_HEADER_INSTANTIATION", "Off")


class CMakeBuild(build_ext):
    def build_extension(self, ext):
        # Must be in this form due to bug in .resolve() only fixed in Python 3.10+
        ext_fullpath = Path.cwd() / self.get_ext_fullpath(ext.name)
        output_directory = ext_fullpath.parent.resolve()
        temp_directory = Path.cwd() / self.build_temp

        print("ext_fullpath", ext_fullpath)
        print("output_directory", output_directory)
        print("temp_directory", temp_directory)

        build_type = "Debug" if os.environ.get('TYR_DEBUG_BUILD') else "Release"
        print("tyr build type:", build_type)

        # Create the temporary build directory, if it does not already exist
        os.makedirs(temp_directory, exist_ok=True)

        #######################################################################
        # Build dependencies
        #######################################################################

        cmake_args = [
            f"-DCMAKE_BUILD_TYPE={build_type}",
            f"-DCMAKE_INSTALL_PREFIX={str(temp_directory / 'dependencies' / 'installs')}",
            f"-DCMAKE_PREFIX_PATH={str(temp_directory / 'dependencies' / 'installs')}",
            f"-DPython_EXECUTABLE={sys.executable}"
        ]

        subprocess.run(
            ["cmake", "-S", f"{str(ext.sourcedir / 'dependencies')}", "-B", f"{str(temp_directory / 'dependencies' / 'build')}"] + cmake_args, cwd=str(temp_directory), check=True
        )

        subprocess.run(
            ["cmake", "--build", f"{str(temp_directory / 'dependencies' / 'build')}", f"-j{get_num_jobs()}"]
        )

        subprocess.run(
            ["cmake", "--install", f"{str(temp_directory / 'dependencies' / 'build')}"]
        )

        shutil.rmtree(f"{str(temp_directory / 'dependencies' / 'build')}")

        #######################################################################
        # Build tyr
        #######################################################################

        cmake_args = [
            f"-DTYR_VERSION_INFO={__version__}",
            f"-DCMAKE_BUILD_TYPE={build_type}",  # not used on MSVC, but no harm
            f"-DCMAKE_PREFIX_PATH={str(temp_directory / 'dependencies' / 'installs')}",
            f"-DPython_EXECUTABLE={sys.executable}",
            "-DBUILD_PYTYR=ON",
            "-DBUILD_TESTS=OFF",
            "-DBUILD_EXECUTABLES=OFF",
            "-DBUILD_PROFILING=OFF",
            "-DTYR_USE_LLD=OFF",  # lld seems to prune code
            f"-DTYR_HEADER_INSTANTIATION={get_header_instantiation()}"
        ]

        subprocess.run(
            ["cmake", "-S", ext.sourcedir, "-B", f"{str(temp_directory / 'build')}"] + cmake_args, cwd=str(temp_directory), check=True
        )

        subprocess.run(
            ["cmake", "--build", f"{str(temp_directory / 'build')}", f"-j{get_num_jobs()}"], cwd=str(temp_directory), check=True
        )

        subprocess.run(
            ["cmake", "--install", f"{str(temp_directory / 'build')}", "--prefix", f"{str(output_directory)}"], check=True
        )

        # Remove unwanted directories.
        unwanted_dirs = ["include", "lib", "lib64"]
        for unwanted_dir in unwanted_dirs:
            dir_path = output_directory / unwanted_dir
            if os.path.exists(dir_path):
                shutil.rmtree(dir_path)
                print(f"Removed {dir_path} from the wheel.")


# The information here can also be placed in setup.cfg - better separation of
# logic and declaration, and simpler if you include description/version in a file.
setup(
    name="pytyr",
    version=__version__,
    author="Dominik Drexler",
    author_email="dominik.drexler@liu.se",
    url="https://github.com/drexlerd/tyr",
    description="Tyr datalog/planning library",
    long_description="",
    install_requires=[],
    packages=find_packages(where="python/src"),
    package_dir={"": "python/src"},
    ext_modules=[CMakeExtension("pytyr")],
    cmdclass={"build_ext": CMakeBuild},
    zip_safe=False,
    extras_require={
        "test": [
            "pytest",
        ],
    }
)



"""
Create planning formalism structures. 

This example demonstrates how to create planning formalism structures.
- Bindings are currently not exhaustively added.
- We currently match the C++ API and work with indices.

Example usage (run from the repository root):

    python3 python/examples/formalism/planning/builder.py \
        -d data/gripper/domain.pddl \
        -p data/gripper/test_problem.pddl

Author: Dominik Drexler (dominik.drexler@liu.se)
"""

import argparse

from pathlib import Path

from pytyr.formalism.planning import ParserOptions, Parser, ObjectBuilder, FluentPredicateBuilder, FluentGroundAtomBuilder


def main():
    arg_parser = argparse.ArgumentParser(description="Create planning formalism structures.")
    arg_parser.add_argument("-d", "--domain-filepath", type=Path, required=True, help="Path to a PDDL domain file.")
    arg_parser.add_argument("-p", "--task-filepath", type=Path, required=True, help="Path to PDDL task file.")
    args = arg_parser.parse_args()

    domain_filepath : Path = args.domain_filepath
    task_filepath : Path = args.task_filepath

    parser_options = ParserOptions()
    parser = Parser(domain_filepath, parser_options)
    
    planning_task = parser.parse_task(task_filepath, parser_options)

    repository = planning_task.get_repository()

    object_builder = ObjectBuilder()
    predicate_builder = FluentPredicateBuilder()
    ground_atom_builder = FluentGroundAtomBuilder()
    
    # 1. Re-create existing atom (at ball1 rooma)
    object_builder.name = "ball1"
    ball1, inserted = repository.get_or_create_object(object_builder)
    assert not inserted
    object_builder.name = "rooma"
    rooma, inserted = repository.get_or_create_object(object_builder)
    assert not inserted 

    predicate_builder.name = "at"
    predicate_builder.arity = 2
    at, inserted = repository.get_or_create_fluent_predicate(predicate_builder)
    assert not inserted

    row, inserted = repository.get_or_create_fluent_predicate_row(at, [ball1.get_index(), rooma.get_index()])
    assert not inserted

    ground_atom_builder.predicate = row.get_index().predicate_index
    ground_atom_builder.row = row.get_index().row_index
    at_ball1_rooma, inserted = repository.get_or_create_fluent_ground_atom(ground_atom_builder)
    assert not inserted

    # 2. Create non existing atom (at ball1 roomb)
    object_builder.name = "roomb"
    roomb, inserted = repository.get_or_create_object(object_builder)
    assert not inserted 

    row, inserted = repository.get_or_create_fluent_predicate_row(at, [ball1.get_index(), roomb.get_index()])
    assert inserted 

    ground_atom_builder.predicate = row.get_index().predicate_index 
    ground_atom_builder.row = row.get_index().row_index
    at_ball1_roomb, inserted = repository.get_or_create_fluent_ground_atom(ground_atom_builder)
    assert inserted

    # 3. Map atom to FDRFact to be passed into state repository for state creations
    fdr_context = planning_task.get_fdr_context()
    fact = fdr_context.get_fact(at_ball1_roomb)

    print(fact)


if __name__ == "__main__":
    main()


"""
Parse and traverse planning formalism structures. 

This example demonstrates how to inspect the planning formalism and highlights
some high-level design choices. The provided type hints make it easy to
navigate and explore all structures.

Example usage (run from the repository root):

    python3 python/examples/formalism/planning/structures.py \
        -d data/gripper/domain.pddl \
        -p data/gripper/test_problem.pddl

Author: Dominik Drexler (dominik.drexler@liu.se)
"""

import argparse

from pathlib import Path

from pytyr.formalism.planning import ParserOptions, Parser


def main():
    arg_parser = argparse.ArgumentParser(description="Parse and traverse planning formalism structures.")
    arg_parser.add_argument("-d", "--domain-filepath", type=Path, required=True, help="Path to a PDDL domain file.")
    arg_parser.add_argument("-p", "--task-filepath", type=Path, required=True, help="Path to PDDL task file.")
    args = arg_parser.parse_args()

    domain_filepath : Path = args.domain_filepath
    task_filepath : Path = args.task_filepath

    parser_options = ParserOptions()
    parser = Parser(domain_filepath, parser_options)
    
    ### 1. Domain: a domain an be shared among arbitrarily many downstream tasks.

    planning_domain = parser.get_domain()
    formalism_domain = planning_domain.get_domain()

    # Tyr separates static/fluent/derived predicates to encode semantic meaning directly within the type.
    print("\nStatic predicates:")
    for predicate in formalism_domain.get_static_predicates():
        print(predicate.get_name(), predicate.get_arity())
    print("\nFluent predicates:")
    for predicate in formalism_domain.get_fluent_predicates():
        print(predicate.get_name(), predicate.get_arity())
    print("\nDerived predicates:")
    for predicate in formalism_domain.get_derived_predicates():
        print(predicate.get_name(), predicate.get_arity())

    # A similar separation exists for functions
    print("\nStatic functions:")
    for function in formalism_domain.get_static_functions():
        print(function.get_name(), function.get_arity())
    print("\nFluent functions:")
    for function in formalism_domain.get_fluent_functions():
        print(function.get_name(), function.get_arity())
    # There is at most one auxiliary function, i.e., "total-cost" whose value is not stored in a state because it does not affect transitions.
    print("\nAuxiliary function:")
    print(formalism_domain.get_auxiliary_function())

    print("\nConstants:")
    for object in formalism_domain.get_constants():
        print(object.get_name())

    ### 2. Task: a task is expected to be defined over the previously parsed domain.

    planning_task = parser.parse_task(task_filepath, parser_options)
    formalism_task = planning_task.get_task()

    # Ensure that the task is defined over the given input domain
    assert(formalism_task.get_domain() == formalism_domain)

    # The separation from above induces a parallel separation for atom, literal, ground atom, and ground literal.
    print("\nStatic initial ground atoms:")
    for ground_atom in formalism_task.get_static_atoms():
        print(ground_atom.get_predicate(), ground_atom.get_objects())
    print("\nFluent initial ground atoms:")
    for ground_atom in formalism_task.get_fluent_atoms():
        print(ground_atom.get_predicate(), ground_atom.get_objects())

    # The task may contain task-specific derived predicates, e.g., miconic-fulladl, 
    # resulting from the translation of complicated goal, i.e., non-conjunctions of literals.
    print("\nTask-specific derived predicates:")
    for predicate in formalism_task.get_derived_predicates():
        print(predicate.get_name(), predicate.get_arity())
    # Similarly, the translation of complicated goals will also introduce task-specific axioms.
    print("\nTask-specific axioms:")
    for axiom in formalism_task.get_axioms():
        print(axiom)


if __name__ == "__main__":
    main()